#!/usr/bin/perl

use strict;
use warnings;

use Date::Parse;	# ..str2time()

use Webutils;

my $cgi = webutils_cgiinit('ARP Watch');
my $dbh = webutils_utminit();

my $script_name = $ENV{SCRIPT_NAME} || $0;
my $macbased = $cgi->param('macbased') || '0';
my $ipaddr   = $cgi->param('ip')  || $cgi->param('ipaddr')  || '';
my $macaddr  = $cgi->param('mac') || $cgi->param('macaddr') || '';
my $summary  = $cgi->param('summary');
my $period   = $cgi->param('period') || 'day';
my $period_seconds = 3600*24;   # ..Day by default

my $logdir = "/var/log/arpwatch";
my $mailbox = "/var/mail/arpwatch";
my  %ip2msglist;
my %mac2msglist;

my %fieldnames = (
	             "Subject" => 'subject',
	          "ip address" => 'ipaddr',
	    "ethernet address" => 'macaddr',
	"old ethernet address" => 'old_macaddr',
	           "timestamp" => 'tstamp',
	  "previous timestamp" => 'previous_tstamp',
	               "delta" => 'delta',
);

$period_seconds = 0       if $period eq 'all';
$period_seconds = 3600    if $period eq 'hour' or $period eq 'lasthour';
$period_seconds = 3600*24 if $period eq 'day'  or $period eq 'lastday';
$period_seconds = $period if $period =~ /^\d+$/ and $period > 0;

$macaddr =~ s/0(\d)/$1/ge;    # ..convert format from common ":07:" to arpwatch ":7:"
$macaddr =~ s/-/:/g;          # ..convert format from dlink
$macaddr = lc($macaddr);

$macbased = 1 if $macaddr and not $ipaddr;
$macbased = 0 if $ipaddr and not $macaddr;

my $macstr = $macaddr ? arpw2macstr($macaddr) : '';
my $now = time;

#======  Parse database  =======================================

if ($dbh) {
	my $q = "select * from arpwatch";
	my @c;
	my @a;
	if ($period_seconds) { push @c, "(tstamp > ? - ?)";                 push @a, $now, $period_seconds }
	if ($macaddr)        { push @c, "(macaddr = ? or old_macaddr = ?)"; push @a, $macaddr, $macaddr }
	if ($ipaddr)         { push @c, "(ipaddr = ?)";                     push @a, $ipaddr }
	$q .= " where ".join(' and ',@c) if @c;

	print "<!-- Macbased = $macbased, MAC = $macstr, Query = \"$q\" -->\n";
	my $sth = execute_query($dbh, $q, @a);
	while (my $r = $sth->fetchrow_hashref) {
		$r->{delta} = ($r->{tstamp} and $r->{previous_tstamp} and $r->{tstamp} > $r->{previous_tstamp})
			? "".($r->{tstamp} - $r->{previous_tstamp})." seconds" : '';
		#print "<!-- tstamp = $r->{tstamp}, prev = $r->{previous_tstamp}, delta = $r->{delta} -->\n";
		$r->{tstamp} = localtime($r->{tstamp});
		$r->{previous_tstamp} = localtime($r->{previous_tstamp})
			if $r->{previous_tstamp};

		my $msgs_ref = $ip2msglist{$r->{ipaddr}};
		$msgs_ref = $ip2msglist{$r->{ipaddr}} = [] unless $msgs_ref;
		push @$msgs_ref, $r;

		$msgs_ref = $mac2msglist{$r->{macaddr}};
		$msgs_ref = $mac2msglist{$r->{macaddr}} = [] unless $msgs_ref;
		push @$msgs_ref, $r;
	}
}


#======  Parse files  ==========================================

sub parse_msgfile($) {
	my $fname = shift;
	unless (open F, $fname) {
		warn "Cannot open $fname: $!\n";
		return;
	}
	my %msg;
	while(<F>) {
		chomp;
		next unless $_;
		next unless /^\s*([^:]+): (.+)$/;
		if ($1 eq 'ip address') {
			my $msgs_ref = $ip2msglist{$2};
			$msgs_ref = $ip2msglist{$2} = [] unless $msgs_ref;
			push @$msgs_ref, \%msg;
		} elsif ($1 eq 'ethernet address') {
			my $msgs_ref = $mac2msglist{$2};
			$msgs_ref = $mac2msglist{$2} = [] unless $msgs_ref;
			push @$msgs_ref, \%msg;
		}
		$msg{$fieldnames{$1} || $1} = $2;
	}
	#while(my ($k,$v) = each %msg) { print "$fname: $k: $v\n" }
	close F;
}

if ($logdir and -r $logdir and -x $logdir) {
opendir(D, $logdir) or die "Cannot open $logdir: $!\n";
while($_ = readdir(D)) {
	#next unless /\.msg/;
	my $mtime = (stat("$logdir/$_"))[9];
	if ($period_seconds and $now - $mtime > $period_seconds) {
		print "<!-- skip $_: mtime = ".localtime($mtime)."-->\n";
		next;
	}
	parse_msgfile "$logdir/$_";
	#last;
}
closedir D;
}


#======  Parse mailbox  ========================================

if ($mailbox and -r $mailbox) {
	open F, $mailbox or die "Cannot open mailbox";
	my $msg;
	my $skip_msg;
	while(<F>) {
		chomp;
		next unless $_;
		
		unless (/^\s*([^:]+): (.+)$/) {
			my @w = split;
			next if $w[0] ne 'From';
			print "\n<!-- $_ -->\n";
			$msg = {};
			# $w[2..6] = 'Wed Apr  9 04:50:57 2008'
			my $tstr = "$w[2] $w[3] $w[4] $w[5] $w[6]";
			my $t = str2time($tstr);
			$skip_msg = ($period_seconds and $now - $t > $period_seconds);
			print "<!-- skip $tstr -->\n" if $skip_msg;
			next;
		}
		next if $skip_msg;
		print "<!-- parse \"$1\" = \"$2\" -->\n";

		if ($1 eq 'ip address') {
			my $msgs_ref = $ip2msglist{$2};
			$msgs_ref = $ip2msglist{$2} = [] unless $msgs_ref;
			push @$msgs_ref, $msg;
		} elsif ($1 eq 'ethernet address') {
			my $msgs_ref = $mac2msglist{$2};
			$msgs_ref = $mac2msglist{$2} = [] unless $msgs_ref;
			push @$msgs_ref, $msg;
		}
		$msg->{$fieldnames{$1} || $1} = $2;
	}
	close F;
}


#======  Filter form  ==========================================

print << "FILTER_FORM";
<form method='get' action='$script_name'>
<input type='hidden' name='period' value='all'/>
<table border='0'>
<tr>
    <td align='right'>IP-address:</td>
    <td><input name='ipaddr' value='$ipaddr'/></td>
    <td align='center' valign='center' rowspan='2'>
	<input type='submit' value='Filter!'/></td>
</tr><tr>
    <td align='right'>MAC-address:</td>
    <td><input name='macaddr' value='$macstr'/></td>
</tr>
</table>
</form>
FILTER_FORM

#======  Navigation header  ====================================

sub print_link($$) {
	my ($val, $name) = @_;
	print "<tr><td align='right'>$name:</td>\n"
	     ."    <td> <a href='$script_name?period=$val'>IP-based</a></td>\n"
	     ."    <td> <a href='$script_name?period=$val&macbased=1'>MAC-based</a></td>"
	     ."    <td> <a href='arpwatch-topstats.cgi?period=$val'>Top stats</a></td>"
	     ."</tr>\n";
}

print "<h3>Display period: $period ($period_seconds seconds), "
	."IP = ".($ipaddr || 'all').", MAC = ".($macstr || 'all')."</h3>\n";

print "<p><table border='0'>\n";
print_link('hour', 'Last hour only');
print_link('day',  'Last day only (default)');
print_link('all',  'All');
print "</table></p>\n";

print "<p><table border='4' cellpadding='6'>\n";
print "<tr><th>IP</th><th>What</th><th>".($macbased ? 'IP' : 'MAC')."</th><th>Old MAC</th>"
	."<th>Time</th><th>Old time</th><th>Delta</th>\n";


#======  Display result  =======================================

my %ips;
my %macs;

sub print_list($$$$) {
	my ($selected, $linkname, $skip_field, $values_ref) = @_;
	while (my ($val, $msgs_ref) = each %$values_ref) {
		next if $selected and $val ne $selected;
		print "<!-- val = $val, selected = $selected -->\n" if $val and $selected;
		next if @$msgs_ref == 1 and $$msgs_ref[0]->{subject} =~ /new station/;
		my $showval = ($linkname eq 'mac' ? arpw2macstr($val) : $val);
		print "<tr><th align='left' colspan='7'>"
			."<a href='$script_name?$linkname=$val&period=all'>$showval</a>"
			."</th></tr>\n";
		foreach my $msg_ref (@$msgs_ref) {
			print "<tr><td></td>";
			foreach(qw/subject ipaddr macaddr old_macaddr tstamp previous_tstamp delta/) {
				my $s = $msg_ref->{$_};
				my $s2 = $s;
				if ($s) {
					++( $ips{$s} ||= 0) if $_ eq 'ipaddr';
					++($macs{$s} ||= 0) if $_ eq 'macaddr' or $_ eq 'old_macaddr';
				}
				next if $_ eq $skip_field;
				if ($s) {
					$s = "<a href='$script_name?period=all&ipaddr=$s' >$s</a>"
						if $_ eq 'ipaddr';
					$s = "<a href='$script_name?period=all&macaddr=$s'>$s2</a>"
						if ($_ eq 'macaddr' or $_ eq 'old_macaddr') and ($s2 = arpw2macstr($s));
				}
				print "<td><!-- field: $_ -->".($s || '&nbsp;')."</td>";
			}
			print "</tr>\n";
		}
	}
}

sub dump_list($$) {
	my ($hint, $hashref) = @_;
	while(my ($k, $v) = each %$hashref) {
		print "$hint = $k\n";
		foreach my $m (@$v) {
			while (my ($k, $v) = each %$m) {
				print "$k:$v  ";
			}
			print "\n";
		}
	}
}

if ($cgi->param('debug')) {
	print "\n<!--\n\n========  MAC List  ========\n\n";
	dump_list "MAC", \%mac2msglist;
	print "\n========  IP List  ========\n\n";
	dump_list "IP",   \%ip2msglist;
	print "\n-->\n\n";
}

if ($macbased or ($macaddr and not $ipaddr)) {
	print_list($macaddr, 'mac', 'macaddr', \%mac2msglist);
} else {
	print_list($ipaddr,  'ip',  'ipaddr',  \%ip2msglist);
}

print "</table></p>";

if ($macaddr or $ipaddr or $summary) {
print << "SUMMARY1";
<p><h3>Summary</h3>
<table>
<tr><td valign='top'>
<table border='1' cellpadding='4'>
<tr><th align='center'>Count</th><th>IP-addrs</th></tr>
SUMMARY1
print "<tr><td align='right'>$ips{$_}</td><td><a href='$script_name?period=all&ipaddr=$_'>$_</a></td></tr>\n"
	foreach sort { $ips{$b} <=> $ips{$a} } keys %ips;
print << "SUMMARY2";
</table>
</td><td>&nbsp;</td><td valign='top'>
<table border='1' cellpadding='4'>
<tr><th align='center'>Count</th><th>MAC-addrs</th></tr>
SUMMARY2
print "<tr><td align='right'>$macs{$_}</td><td><a href='$script_name?period=all&macaddr=$_'>$_</a></td></tr>\n"
	foreach sort { $macs{$b} <=> $macs{$a} } keys %macs;
print "</table>\n";
print "</td></tr></table></p>\n";
}

#$dbh->disconnect;

webutils_cgidone($cgi);

## EOF ##
